----------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand 
-- 
-- Create Date:    10:28:58 11/12/2015 
-- Design Name: 
-- Module Name:    Datapath - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

LIBRARY work;
USE work.API_pkg.ALL;

entity Datapath is
  port(
    clk_i  : in std_logic;
    --Input
    data_i    : in std_logic_vector(3 downto 0);
    key_i     : in std_logic_vector(3 downto 0);
    --Output
    data_o    : out std_logic_vector(3 downto 0);
    --Control signals
    key_sel   : in std_logic_vector(1 downto 0);
    data_sel  : in std_logic_vector(2 downto 0);
    mode_i    : in std_logic;
    rc_init   : in std_logic_vector(1 downto 0);
    round_sel : in std_logic;
    count_r   : in std_logic_vector(1 downto 0);
    int_count : in std_logic_vector(3 downto 0)
  );
end Datapath;

architecture Behavioral of Datapath is

  COMPONENT generate_constant_led
	PORT(
		rc_in : IN std_logic_vector(5 downto 0);
		mode_i : IN std_logic;          
		rc_out : OUT std_logic_vector(5 downto 0)
		);
	END COMPONENT;
  
  COMPONENT sbox
	PORT(
		sbox_i : IN std_logic_vector(3 downto 0);          
		sbox_o : OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;
  
  COMPONENT sbox_d
	PORT(
		sbox_i : IN std_logic_vector(3 downto 0);          
		sbox_o : OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;
  
  COMPONENT mix_nibble
	PORT(
		word_i : IN std_logic_vector(15 downto 0);          
		nible_o : OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;
  
  COMPONENT mix_nibled
	PORT(
		word_i : IN std_logic_vector(15 downto 0);          
		nible_o : OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;
  
  CONSTANT ks : STD_LOGIC_VECTOR(7 DOWNTO 0) := x"80"; --128
  --regsiters
  signal key_reg      : std_logic_vector(key_size-1 downto 0);
  signal state_reg    : std_logic_vector(state_size-1 downto 0);
  signal rc_reg       : std_logic_vector(5 downto 0);
  --signals
  signal add_rk_i    : std_logic_vector(3 downto 0);
  signal add_rk_o    : std_logic_vector(3 downto 0);
  signal sub_i       : std_logic_vector(3 downto 0);
  signal sub_o       : std_logic_vector(3 downto 0);
  signal subd_o      : std_logic_vector(3 downto 0);
  signal round_1_o   : std_logic_vector(3 downto 0);
  signal round_2_o   : std_logic_vector(3 downto 0);
  signal round_o     : std_logic_vector(3 downto 0);
  signal rc_o        : std_logic_vector(5 downto 0);
  signal add_const_i : std_logic_vector(3 downto 0);
  signal conste      : std_logic_vector(3 downto 0);
  signal constd      : std_logic_vector(3 downto 0);
  signal const       : std_logic_vector(3 downto 0);
  signal mix_i       : std_logic_vector(15 downto 0);
  signal mixd_i       : std_logic_vector(15 downto 0);
  signal mix_o       : std_logic_vector(3 downto 0);
  signal mixd_o      : std_logic_vector(3 downto 0);

begin

  --key process
  K_proc : process(clk_i, key_sel)
  begin
    if rising_edge(clk_i) then
      if key_sel = "01" then
        key_reg <= key_i & key_reg(key_size-1 downto 4);
      elsif key_sel = "10" then
        if mode_i = '0' then
          key_reg <= key_reg(3 downto 0) & key_reg(key_size-1 downto 4);
        else
          key_reg <= key_reg(key_size-5 downto 0) & key_reg(key_size-1 downto key_size-4);
        end if;
      elsif key_sel = "11" then
        key_reg <= key_reg(key_size-5 downto 0) & key_reg(key_size-1 downto key_size-4);
      else
        key_reg <= key_reg;
      end if;
    end if;
  end process;
  
  --data process
  D_proc : process(clk_i, data_sel)
  begin
    if rising_edge(clk_i) then
      if data_sel = "001" then
        state_reg <= data_i & state_reg(state_size-1 downto 4);
      elsif data_sel = "101" then
        state_reg <= (data_i xor key_reg(63 downto 60)) & state_reg(state_size-1 downto 4);
      elsif data_sel = "010" then
        state_reg <=  round_o & state_reg(63 downto 4);
      elsif data_sel = "011" then
        if mode_i = '0' then
          state_reg(63 downto 48) <= state_reg(63 downto 52) & round_o;
          state_reg(47 downto 32) <= state_reg(43 downto 36) & state_reg(51 downto 44);
          state_reg(31 downto 16) <= state_reg(23 downto 20) & state_reg(35 downto 24);
          state_reg(15 downto 0)  <= state_reg(19 downto 4);
        else
          state_reg(63 downto 48) <=  round_o & state_reg(63 downto 52);
          state_reg(47 downto 32) <= state_reg(39 downto 36) & state_reg(51 downto 40);
          state_reg(31 downto 16) <= state_reg(27 downto 20) & state_reg(35 downto 28);
          state_reg(15 downto 0)  <= state_reg(15 downto 4) & state_reg(19 downto 16);
        end if;
      elsif data_sel = "111" then
        state_reg <= add_rk_o & state_reg(63 downto 4);
      else
        state_reg <= state_reg;
      end if;
    end if;
  end process;
  
  Inst_generate_constant_led: generate_constant_led PORT MAP(
		rc_in => rc_reg,
		mode_i => mode_i,
		rc_out => rc_o
	); 
  
  rc_proc : process(clk_i, rc_init)
  begin
    if rising_edge(clk_i) then
      if rc_init = "01" then
        rc_reg  <= "000001";
      elsif rc_init = "10" then
        rc_reg  <= "000100";
      elsif rc_init = "11" then
        rc_reg  <= rc_o;
      else
        rc_reg  <= rc_reg;
      end if;
    end if;
  end process;
  
  data_o <= state_reg(3 downto 0);
  
  --addrk
  with mode_i select
    add_rk_i <=
      state_reg(3 downto 0) when '0',
      sub_i when others;
  
  with (mode_i & count_r) select
    add_rk_o <= 
      add_rk_i xor key_reg(3 downto 0) when "000",
      add_rk_i xor key_reg(63 downto 60) when "111",
      add_rk_i when others;
  
  with int_count(3 downto 0) select
    conste <= 
      ks(7 downto 4) when "0000",
      x"1" xor ks(7 downto 4) when "0100",
      x"2" xor ks(3 downto 0) when "1000",
      x"3" xor ks(3 downto 0) when "1100",
      '0' & rc_reg(5 downto 3) when "0001",
      '0' & rc_reg(2 downto 0) when "0101",
      '0' & rc_reg(5 downto 3) when "1001",
      '0' & rc_reg(2 downto 0) when "1101",
      "0000" when others;
  
  with int_count(3 downto 0) select
    constd <= 
      ks(7 downto 4) when "1111",
      x"1" xor ks(7 downto 4) when "1011",
      x"2" xor ks(3 downto 0) when "0111",
      x"3" xor ks(3 downto 0) when "0011",
      '0' & rc_reg(5 downto 3) when "1110",
      '0' & rc_reg(2 downto 0) when "1010",
      '0' & rc_reg(5 downto 3) when "0110",
      '0' & rc_reg(2 downto 0) when "0010",
      "0000" when others;
      
   with mode_i select 
    const <= 
      conste when '0',
      constd when others;
  
  Inst_sbox_d: sbox_d PORT MAP(
		sbox_i => state_reg(3 downto 0),
		sbox_o => subd_o
	);
  
  with mode_i select
    add_const_i <=
      add_rk_o when '0',
      subd_o when others;
    
  sub_i <= add_const_i xor const;
  
  Inst_sbox : sbox port map(
    sbox_i => sub_i,
    sbox_o => sub_o
  );
  
  with mode_i select 
    round_1_o <= 
      sub_o when '0',
      mixd_o when others;
      
  with mode_i select 
    round_2_o <= 
      mix_o when '0',
      add_rk_o when others;
  
  with round_sel select
    round_o <=
      round_1_o when '1',
      round_2_o when others;
      
  mix_i <= state_reg(51 downto 48) & 
           state_reg(35 downto 32) & 
           state_reg(19 downto 16) & 
           state_reg(3 downto 0);
  
  Inst_mix_nibble: mix_nibble PORT MAP(
		word_i => mix_i,
		nible_o => mix_o
	);
  
  Inst_mix_nibled: mix_nibled PORT MAP(
		word_i => mix_i,
		nible_o => mixd_o
	);
  
end Behavioral;

