---------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    10:50:44 10/29/2015 
-- Design Name: 
-- Module Name:    Top_level - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

LIBRARY work;
USE work.API_pkg.ALL;

entity Top_level is
  port(
    rstb_i    : in std_logic;
    clk_i     : in std_logic;
    --input
    control_i : in std_logic_vector(1 downto 0);
    data_i    : in std_logic_vector(3 downto 0);
    key_i     : in std_logic_vector(3 downto 0);
    --output
    busy_o    : out std_logic;
    key_valid : out std_logic;
    done_o    : out std_logic;
    data_o    : out std_logic_vector(3 downto 0)
  );
end Top_level;

architecture Behavioral of Top_level is

  --Control unit
  COMPONENT control_unit
	generic (
    state_size  : in integer := 64;
    key_size    : in integer := 80
  );
  port(
    clk_i     : in std_logic;
    rstb_i    : in std_logic;
    --instruction
    control_i : in std_logic_vector(1 downto 0);
    --output flag
    key_ready : out std_logic;
    busy_o    : out std_logic;
    done_o    : out std_logic;
    --control signal
    mode_o    : out std_logic;
    key_sel   : out std_logic_vector(1 downto 0);
    data_sel  : out std_logic_vector(2 downto 0);
    rc_init   : out std_logic_vector(1 downto 0);
    round_sel : out std_logic;
    count_r   : out std_logic_vector(1 downto 0);
    int_count : out std_logic_vector(3 downto 0)
  );
	END COMPONENT;
  
  COMPONENT Datapath
	port(
    clk_i  : in std_logic;
    --Input
    data_i    : in std_logic_vector(3 downto 0);
    key_i     : in std_logic_vector(3 downto 0);
    --Output
    data_o    : out std_logic_vector(3 downto 0);
    --Control signals
    key_sel   : in std_logic_vector(1 downto 0);
    data_sel  : in std_logic_vector(2 downto 0);
    mode_i    : in std_logic;
    rc_init   : in std_logic_vector(1 downto 0);
    round_sel : in std_logic;
    count_r   : in std_logic_vector(1 downto 0);
    int_count : in std_logic_vector(3 downto 0)
  );
	END COMPONENT;
  --signals
  --regsiters
--  signal key_reg      : std_logic_vector(key_size-1 downto 0);
--  signal state_reg    : std_logic_vector(state_size-1 downto 0);
  signal rc_sel       : std_logic_vector(1 downto 0);
  signal round_sel_s  : std_logic;
  
  --internal control signals
  signal key_sel_s    : std_logic_vector(1 downto 0);
  signal data_sel_s   : std_logic_vector(2 downto 0);
  signal round_c_s    : std_logic_vector(1 downto 0);
  signal int_count_s  : std_logic_vector(3 downto 0);
  signal mode_s       : std_logic;
  
begin

  Inst_control_unit: control_unit 
  GENERIC MAP(
    key_size => 128
  )
  PORT MAP(
		clk_i     => clk_i,
		rstb_i    => rstb_i,
		control_i => control_i,
		key_ready => key_valid,
		busy_o    => busy_o,
		done_o    => done_o,
		mode_o    => mode_s,
    key_sel   => key_sel_s,
    data_sel  => data_sel_s,
    rc_init   => rc_sel,
    round_sel => round_sel_s,
    count_r   => round_c_s,
    int_count => int_count_s
	);
  
  Inst_Datapath: Datapath PORT MAP(
		clk_i => clk_i,
		data_i => data_i,
		key_i => key_i,
		data_o => data_o,
		key_sel => key_sel_s,
		data_sel => data_sel_s,
    mode_i    => mode_s,
    rc_init => rc_sel,
    round_sel => round_sel_s,
		count_r => round_c_s(1 downto 0),
    int_count => int_count_s
	);
  
end Behavioral;

